#include "system.h"

static Texture* clOutput = 0;
static Shader* shader = 0;
static Kernel* kernel = 0;
static Buffer* outputBuffer, *logBuffer;

bool Game::Init()
{
	// load shader and texture
	clOutput = new Texture( SCRWIDTH, SCRHEIGHT, Texture::FLOAT );
	shader = new Shader( "shaders/vignette.vert", "shaders/vignette.frag" );
	// load OpenCL code
	kernel = new Kernel( "programs/program.cl", "task" );
	// link cl output texture as an OpenCL buffer
	outputBuffer = new Buffer( clOutput->GetID(), Buffer::TEXTURE );
	kernel->SetArgument( 0, outputBuffer );
	kernel->SetArgument( 1, Kernel::GetLogBuffer() );
	// done
	return true;
}

void Game::Tick()
{
	// run cl code to fill texture
	kernel->Run( outputBuffer );
	// run shader on cl-generated texture
	shader->Bind();
	shader->SetInputTexture( GL_TEXTURE0, "color", clOutput );
#ifdef USE_MATHFU
	shader->SetInputMatrix( "view", mat4::Identity() );
#else
	shader->SetInputMatrix( "view", mat4( 1 ) );
#endif
	DrawQuad();
}

void Game::Shutdown()
{
	delete clOutput;
	delete shader;
	delete kernel;
}

// EOF