#include "../shared.h"
#include "program.h"
#include "tools/logging.h"
#include "tools/random.h"

void intersectPlane( float3 O, float3 D, float3 N, float d, float* dist )
{
	float t = -(dot( O, N ) + d) / dot( D, N );
	if (t > 0) *dist = min( *dist, t );
}

void intersectSphere( float3 O, float3 D, float3 pos, float r, float* dist )
{
	float3 p = O - pos;
	float b = -dot( p, D );
	float det = b * b - dot( p, p ) + r * r;
	if (det < 0) return;
	det = sqrt( det );
	float t1 = b - det, t2 = b + det;
	if (t2 < 0) return;
	if (t1 < 0) t1=0;
	*dist = min( *dist, min( t2, t1 ) );
}

void intersectScene( float3 O, float3 D, float* dist )
{
	float dummy;
	float x = fract( 2.5, &dummy );
	*dist = 10000;
	intersectPlane( O, D, (float3)( 0, 1, 0 ), -1.0, dist );
	intersectSphere( O, D, (float3)( -0.7, 0, -2 ), 0.3, dist );
	intersectSphere( O, D, (float3)( 0, 0, -2 ), 0.3, dist );
	intersectSphere( O, D, (float3)( 0.7, 0, -2 ), 0.3, dist );
}

__kernel void task( write_only image2d_t outimg, __global uint* logBuffer )
{
	int column = get_global_id( 0 );
	int line = get_global_id( 1 );
	float2 uv = { (float)column / 800, (float)line / 480 };
	// screen corners
	float3 p1 = { -1, -0.7, -1.5 }, p2 = { 1, -0.7, -1.5 }, p4 = { -1, 0.7, -1.5 };
	// calculate pixel color with anti-aliasing
	float color = 0;
	for( int x = 0; x < 4; x++ ) for( int y = 0; y < 4; y++ )
	{
		// generate primary ray
		float3 P = p1 + (uv.x + (float)x * (0.25 / 800)) * (p2 - p1) + (uv.y + (float)y * (0.25 / 480)) * (p4 - p1);
		float3 O = { 0, 0, 0 };
		float3 D = normalize( P - O );
		// intersect primitives
		float t;
		intersectScene( O, D, &t );
		// sample color
		float c = max( 0.0, 10.0 - t );
		// accumulate
		color += c;
	}
	// visualize
	write_imagef( outimg, (int2)(column, line), 0.1f * (color / 16) );
}