bool IntersectBox( const float3* bmin, const float3* bmax, const float3* O, const float3* rD, float* dist )
{
	const float3 b1 = (*bmin - *O) * *rD, b2 = (*bmax - *O) * *rD;
	float nr = min( b1.x, b2.x ), fr = max( b1.x, b2.x );
	float tmin = (0.0f < nr) ? nr : 0.0f, tmax = (fr <  *dist) ? fr : *dist;
	nr = min( b1.y, b2.y ), fr = max( b1.y, b2.y );
	tmin = (tmin < nr) ? nr : tmin, tmax = (fr <  tmax) ? fr : tmax;
	nr = min( b1.z, b2.z ), fr = max( b1.z, b2.z );
	tmin = (tmin < nr) ? nr : tmin, tmax = (fr <  tmax) ? fr : tmax;
	if ((tmin > tmax) || (tmin > *dist)) return false;
	*dist = tmin;
	return true;
}

bool CheckBox( const float3* bmin, const float3* bmax, const float3* O, const float3* rD, const float* dist )
{
	const float3 b1 = (*bmin - *O) * *rD, b2 = (*bmax - *O) * *rD;
	float nr = min( b1.x, b2.x ), fr = max( b1.x, b2.x );
	float tmin = (0.0f < nr) ? nr : 0.0f, tmax = (fr <  *dist) ? fr : *dist;
	nr = min( b1.y, b2.y ), fr = max( b1.y, b2.y );
	tmin = (tmin < nr) ? nr : tmin, tmax = (fr <  tmax) ? fr : tmax;
	nr = min( b1.z, b2.z ), fr = max( b1.z, b2.z );
	tmin = (tmin < nr) ? nr : tmin, tmax = (fr <  tmax) ? fr : tmax;
	return ((tmin <= tmax) && (tmin < *dist));
}

bool IntersectDSSphere( const float3* C, const float* r, const float3* O, const float3* rD, float* dist )
{
	const float3 p = *C - *O;
	const float q = dot( p, *rD );
	const float d = (q * q) - dot( p, p ) + *r * *r;
	const float sqd = sqrt( d );
	float t1 = q - sqd, t2 = q + sqd;
	t1 = (t1 >= 0) ? t1 : 1e30f, t2 = (t2 >= 0) ? t2 : 1e30f;
	float t = min( t1, t2 );
	if ((d <= 0) || (t >= *dist)) return false;
	*dist = t;
	return true;
}

bool CheckDSSphere( const float3* C, const float* r, const float3* O, const float3* rD, const float* dist )
{
	const float3 p = *C - *O;
	const float q = dot( p, *rD );
	const float d = (q * q) - dot( p, p ) + *r * *r;
	const float sqd = sqrt( d );
	float t1 = q - sqd, t2 = q + sqd;
	t1 = (t1 >= 0) ? t1 : 1e30f, t2 = (t2 >= 0) ? t2 : 1e30f;
	float t = min( t1, t2 );
	return ((d > 0) && (t < *dist));
}

bool IntersectSSSphere( const float3* C, const float* r, const float3* O, const float3* D, float* dist )
{
	const float3 c = *C - *O;
	const float t = dot( c, *D ), p = length( c - t * *D );
	if (p > *r) return false; else return t - sqrt( *r * *r - p * p );
}

bool CheckPlane( const float3* N, const float* d, const float3* O, const float3* D, const float* dist )
{
	float v = dot( *D, *N );
	if (v == 0) return false;
	if ((-(dot( *O, *N ) + *d) / v) >= *dist) return false;
	return true;
}

bool IntersectPlane( const float3* N, const float* d, const float3* O, const float3* D, float* dist )
{
	float v = dot( *D, *N );
	if (v == 0) return false;
	float t = -(dot( *O, *N ) + *d) / v;
	if (t > *dist) return false;
	*dist = t;
	return true;
}

bool IntersectTriangle( const float3* v0, const float3* v1, const float3* v2, const float3* O, const float3* D, float* dist )
{
	const float3 e1 = *v1 - *v0, e2 = *v2 - *v0;
	const float3 pvec = cross( *D, e2 );
	const float det = dot( e1, pvec );
	if (det < -EPSILON) return false;
	const float3 tvec = *O - *v0;
	const float u = dot( tvec, pvec );
	if (u < 0.0f || u > det) return false;
	const float3 qvec = cross( tvec, e1 );
	float v = dot( *D, qvec );
	if (v < 0.0f || u + v > det) return false;
	float inv_det = 1.0f / det;
	float t = dot( e2, qvec ) * inv_det;
	if (t >= *dist) return false;
	*dist = t;
	// *out_u = u * inv_det;
	// *out_v = v * inv_det;
	return true;
}