#ifdef LOGGING_ENABLED

#define LogMarker( id, lb ) _LogMarker( id, lb )
#define LogInt( id, v, lb ) _LogInt( id, v, lb )
#define LogFloat( id, v, lb ) _LogFloat( id, v, lb )
#define LogFloat3( id, v, lb ) _LogFloat3( id, v, lb )
#define Error_( f, l, lb ) { _ErrorMarker( f, l, lb ); return; }
#define Assert_( f, l, b, lb ) _AssertMarker( f, l, b, lb )

#else

#define LogMarker( id, lb ) 
#define LogInt( id, v, lb ) 
#define LogFloat( id, v, lb ) 
#define LogFloat3( id, v, lb ) 
#define Error_( f, l, lb )
#define Assert_( f, l, b, lb )

#endif

void _AssertMarker( uint fileNr, uint lineNr, bool a, __global uint* logBuffer )
{
	if (!a)
	{
		uint count = atomic_inc( logBuffer );
		if (count < MAXLOGSIZE)
		{
			logBuffer[count * 2 + 2] = ASSERTFAILED;
			logBuffer[count * 2 + 3] = (lineNr << 8) + fileNr;
		}
	}
}

void _ErrorMarker( uint fileNr, uint lineNr, __global uint* logBuffer )
{
	uint count = atomic_inc( logBuffer );
	if (count < MAXLOGSIZE)
	{
		logBuffer[count * 2 + 2] = ERRORCODE;
		logBuffer[count * 2 + 3] = (lineNr << 8) + fileNr;
	}
}

void _LogMarker( uint id, __global uint* logBuffer )
{
	uint count = atomic_inc( logBuffer );
	if (count < MAXLOGSIZE)
	{
		logBuffer[count * 2 + 2] = id;
		logBuffer[count * 2 + 3] = 0;
	}
}

void _LogInt( uint id, int value, __global uint* logBuffer )
{
	uint count = atomic_inc( logBuffer );
	if (count < MAXLOGSIZE)
	{
		logBuffer[count * 2 + 2] = id;
		logBuffer[count * 2 + 3] = *(uint*)&value;
	}
}

void _LogFloat( uint id, float value, __global uint* logBuffer )
{
	uint count = atomic_inc( logBuffer );
	if (count < MAXLOGSIZE)
	{
		logBuffer[count * 2 + 2] = id;
		logBuffer[count * 2 + 3] = *(uint*)&value;
	}
}

void _LogFloat3( uint id, float3 value, __global uint* logBuffer )
{
	uint count = atomic_add( logBuffer, 2 );
	if (count < (MAXLOGSIZE - 1))
	{
		logBuffer[count * 2 + 2] = id;
		float x = value.x, y = value.y, z = value.z;
		logBuffer[count * 2 + 3] = *(uint*)&x;
		logBuffer[count * 2 + 4] = *(uint*)&y;
		logBuffer[count * 2 + 5] = *(uint*)&z;
	}
}