#pragma once

#define USE_GLEE		// instead of glew; potentially faster but not as safe
#define USE_MATHFU		// insread of glm

#define FATALERROR(m) FatalError( __FILE__, __LINE__, m )
#define ERRORMESSAGE(m,c) FatalError( __FILE__, __LINE__, m, c )
#define CHECKCL(r) CheckCL( r, __FILE__, __LINE__ )

#define GLM_FORCE_RADIANS

// system includes
#include <cl/cl.h>
#include <cl/cl_gl_ext.h>
#ifdef USE_GLEE
#include <GL/glee.h>
#else
#include <GL/glew.h>
#include <GL/wglew.h>
#endif
#include <vector>
#ifdef USE_MATHFU
#include "mathfu/glsl_mappings.h"
#else
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#endif
#include <fstream>
#include <fcntl.h>
#include <io.h>
#include <string>

// external libraries
#include "freeimage.h"

// namespaces
using namespace std;
#ifdef USE_MATHFU
using namespace mathfu;
typedef unsigned int uint;
#else
using namespace glm;
#endif

// framework includes
#include "shared.h"
#include "gltools.h"
#include "opencl.h"
#include "shader.h"
#include "texture.h"
#include "game.h"

// library namespace
using namespace ocllab;

// global helper functions
HWND GetWindowHandle();
void FatalError( const char* file, int line, const char* message );
void FatalError( const char* file, int line, const char* message, const char* context );
float GetTime();
void StartTimer();
float GetDuration();

// random numbers
inline float Rand( float range ) { return ((float)rand() / RAND_MAX) * range; }
inline int IRand( int range ) { return rand() % range; }

// min / max
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

// PI
#define PI	3.14159265358979323846264f

// EOF